import pytest

from vyper.ast.identifiers import RESERVED_KEYWORDS
from vyper.builtins.functions import BUILTIN_FUNCTIONS
from vyper.codegen.expr import ENVIRONMENT_VARIABLES
from vyper.exceptions import NamespaceCollision, StructureException, SyntaxException
from vyper.semantics.types.primitives import AddressT

ALL_RESERVED_KEYWORDS = BUILTIN_FUNCTIONS | RESERVED_KEYWORDS | ENVIRONMENT_VARIABLES


@pytest.mark.parametrize("constant", sorted(ALL_RESERVED_KEYWORDS))
def test_reserved_keywords_memory(constant, get_contract, assert_compile_failed):
    code = f"""
@external
def test():
    {constant}: int128 = 31337
    """
    assert_compile_failed(
        lambda: get_contract(code), (SyntaxException, StructureException, NamespaceCollision)
    )


@pytest.mark.parametrize("constant", sorted(ALL_RESERVED_KEYWORDS))
def test_reserved_keywords_storage(constant, get_contract, assert_compile_failed):
    code = f"{constant}: int128"
    assert_compile_failed(
        lambda: get_contract(code), (SyntaxException, StructureException, NamespaceCollision)
    )


@pytest.mark.parametrize("constant", sorted(ALL_RESERVED_KEYWORDS))
def test_reserved_keywords_fn_args(constant, get_contract, assert_compile_failed):
    code = f"""
@external
def test({constant}: int128):
    pass
    """
    assert_compile_failed(
        lambda: get_contract(code), (SyntaxException, StructureException, NamespaceCollision)
    )


SELF_NAMESPACE_MEMBERS = set(AddressT._type_members.keys())
DISALLOWED_FN_NAMES = SELF_NAMESPACE_MEMBERS | RESERVED_KEYWORDS
ALLOWED_FN_NAMES = ALL_RESERVED_KEYWORDS - DISALLOWED_FN_NAMES


@pytest.mark.parametrize("constant", sorted(ALLOWED_FN_NAMES))
def test_reserved_keywords_fns_pass(constant, get_contract, assert_compile_failed):
    code = f"""
@external
def {constant}(var: int128):
    pass
    """
    assert get_contract(code) is not None


@pytest.mark.parametrize("constant", sorted(DISALLOWED_FN_NAMES))
def test_reserved_keywords_fns_fail(constant, get_contract, assert_compile_failed):
    code = f"""
@external
def {constant}(var: int128):
    pass
    """
    assert_compile_failed(
        lambda: get_contract(code), (SyntaxException, StructureException, NamespaceCollision)
    )



import pytest

from vyper.exceptions import ImmutableViolation, InvalidType, TypeMismatch


def test_augassign(get_contract):
    augassign_test = """
@external
def augadd(x: int128, y: int128) -> int128:
    z: int128 = x
    z += y
    return z

@external
def augmul(x: int128, y: int128) -> int128:
    z: int128 = x
    z *= y
    return z

@external
def augsub(x: int128, y: int128) -> int128:
    z: int128 = x
    z -= y
    return z

@external
def augmod(x: int128, y: int128) -> int128:
    z: int128 = x
    z %= y
    return z
    """

    c = get_contract(augassign_test)

    assert c.augadd(5, 12) == 17
    assert c.augmul(5, 12) == 60
    assert c.augsub(5, 12) == -7
    assert c.augmod(5, 12) == 5
    print("Passed aug-assignment test")


@pytest.mark.parametrize(
    "typ,in_val,out_val",
    [
        ("uint256", 77, 123),
        ("uint256[3]", [1, 2, 3], [4, 5, 6]),
        ("DynArray[uint256, 3]", [1, 2, 3], [4, 5, 6]),
        ("Bytes[5]", b"vyper", b"conda"),
    ],
)
def test_internal_assign(get_contract, typ, in_val, out_val):
    code = f"""
@internal
def foo(x: {typ}) -> {typ}:
    x = {out_val}
    return x

@external
def bar(x: {typ}) -> {typ}:
    return self.foo(x)
    """
    c = get_contract(code)

    assert c.bar(in_val) == out_val


def test_internal_assign_struct(get_contract):
    code = """
flag Bar:
    BAD
    BAK
    BAZ

struct Foo:
    a: uint256
    b: DynArray[Bar, 3]
    c: String[5]

@internal
def foo(x: Foo) -> Foo:
    x = Foo(a=789, b=[Bar.BAZ, Bar.BAK, Bar.BAD], c=\"conda\")
    return x

@external
def bar(x: Foo) -> Foo:
    return self.foo(x)
    """
    c = get_contract(code)

    assert c.bar((123, [1, 2, 4], "vyper")) == (789, [4, 2, 1], "conda")


def test_internal_assign_struct_member(get_contract):
    code = """
flag Bar:
    BAD
    BAK
    BAZ

struct Foo:
    a: uint256
    b: DynArray[Bar, 3]
    c: String[5]

@internal
def foo(x: Foo) -> Foo:
    x.a = 789
    x.b.pop()
    return x

@external
def bar(x: Foo) -> Foo:
    return self.foo(x)
    """
    c = get_contract(code)

    assert c.bar((123, [1, 2, 4], "vyper")) == (789, [1, 2], "vyper")


def test_internal_augassign(get_contract):
    code = """
@internal
def foo(x: int128) -> int128:
    x += 77
    return x

@external
def bar(x: int128) -> int128:
    return self.foo(x)
    """
    c = get_contract(code)

    assert c.bar(123) == 200


@pytest.mark.parametrize("typ", ["DynArray[uint256, 3]", "uint256[3]"])
def test_internal_augassign_arrays(get_contract, typ):
    code = f"""
@internal
def foo(x: {typ}) -> {typ}:
    x[1] += 77
    return x

@external
def bar(x: {typ}) -> {typ}:
    return self.foo(x)
    """
    c = get_contract(code)

    assert c.bar([1, 2, 3]) == [1, 79, 3]


def test_invalid_external_assign(assert_compile_failed, get_contract):
    code = """
@external
def foo(x: int128):
    x = 5
"""
    assert_compile_failed(lambda: get_contract(code), ImmutableViolation)


def test_invalid_external_augassign(assert_compile_failed, get_contract):
    code = """
@external
def foo(x: int128):
    x += 5
"""
    assert_compile_failed(lambda: get_contract(code), ImmutableViolation)


def test_valid_literal_increment(get_contract):
    code = """
storx: uint256

@external
def foo1() -> int128:
    x: int128 = 122
    x += 1
    return x

@external
def foo2() -> uint256:
    x: uint256 = 122
    x += 1
    return x

@external
def foo3(y: uint256) -> uint256:
    self.storx = y
    self.storx += 1
    return self.storx
"""
    c = get_contract(code)

    assert c.foo1() == 123
    assert c.foo2() == 123
    assert c.foo3(11) == 12


def test_invalid_uint256_assignment(assert_compile_failed, get_contract):
    code = """
storx: uint256

@external
def foo2() -> uint256:
    x: uint256 = -1
    x += 1
    return x
"""
    assert_compile_failed(lambda: get_contract(code), TypeMismatch)


def test_invalid_uint256_assignment_calculate_literals(get_contract):
    code = """
storx: uint256

@external
def foo2() -> uint256:
    x: uint256 = 0
    x = 3 * 4 // 2 + 1 - 2
    return x
"""
    c = get_contract(code)

    assert c.foo2() == 5


# See #838. Confirm that nested keys and structs work properly.
def test_nested_map_key_works(get_contract):
    code = """
struct X:
    a: int128
    b: int128
struct Y:
    c: int128
    d: int128
test_map1: HashMap[int128, X]
test_map2: HashMap[int128, Y]

@external
def set():
    self.test_map1[1].a = 333
    self.test_map2[333].c = 111


@external
def get(i: int128) -> int128:
    idx: int128 = self.test_map1[i].a
    return self.test_map2[idx].c
    """
    c = get_contract(code)
    c.set()
    assert c.get(1) == 111


def test_nested_map_key_problem(get_contract):
    code = """
struct X:
    a: int128
    b: int128
struct Y:
    c: int128
    d: int128
test_map1: HashMap[int128, X]
test_map2: HashMap[int128, Y]

@external
def set():
    self.test_map1[1].a = 333
    self.test_map2[333].c = 111


@external
def get() -> int128:
    return self.test_map2[self.test_map1[1].a].c
    """
    c = get_contract(code)
    c.set()
    assert c.get() == 111


@pytest.mark.parametrize(
    "contract",
    [
        """
@external
def foo():
    y: int128 = 1
    z: decimal = y
    """,
        """
@external
def foo():
    y: int128 = 1
    z: decimal = 0.0
    z = y
    """,
        """
@external
def foo():
    y: bool = False
    z: decimal = y
    """,
        """
@external
def foo():
    y: bool = False
    z: decimal = 0.0
    z = y
    """,
        """
@external
def foo():
    y: uint256 = 1
    z: int128 = y
    """,
        """
@external
def foo():
    y: uint256 = 1
    z: int128 = 0
    z = y
    """,
        """
@external
def foo():
    y: int128 = 1
    z: bytes32 = y
    """,
        """
@external
def foo():
    y: int128 = 1
    z: bytes32 = empty(bytes32)
    z = y
    """,
        """
@external
def foo():
    y: uint256 = 1
    z: bytes32 = y
    """,
        """
@external
def foo():
    y: uint256 = 1
    z: bytes32 = empty(bytes32)
    z = y
    """,
    ],
)
def test_invalid_implicit_conversions(contract, assert_compile_failed, get_contract):
    assert_compile_failed(lambda: get_contract(contract), TypeMismatch)


def test_invalid_nonetype_assignment(assert_compile_failed, get_contract):
    code = """
@internal
def bar():
    pass

@external
def foo():
    ret : bool = self.bar()
"""
    assert_compile_failed(lambda: get_contract(code), InvalidType)

    # GH issue 2418


overlap_codes = [
    """
@external
def bug(xs: uint256[2]) -> uint256[2]:
    # Initial value
    ys: uint256[2] = xs
    ys = [ys[1], ys[0]]
    return ys
    """,
    """
foo: uint256[2]
@external
def bug(xs: uint256[2]) -> uint256[2]:
    # Initial value
    self.foo = xs
    self.foo = [self.foo[1], self.foo[0]]
    return self.foo
    """,
    # TODO add transient tests when it's available
]


@pytest.mark.parametrize("code", overlap_codes)
def test_assign_rhs_lhs_overlap(get_contract, code):
    c = get_contract(code)

    assert c.bug([1, 2]) == [2, 1]


def test_assign_rhs_lhs_partial_overlap(get_contract):
    # GH issue 2418, generalize when lhs is not only dependency of rhs.
    code = """
@external
def bug(xs: uint256[2]) -> uint256[2]:
    # Initial value
    ys: uint256[2] = xs
    ys = [xs[1], ys[0]]
    return ys
    """
    c = get_contract(code)

    assert c.bug([1, 2]) == [2, 1]


def test_assign_rhs_lhs_overlap_dynarray(get_contract):
    # GH issue 2418, generalize to dynarrays
    code = """
@external
def bug(xs: DynArray[uint256, 2]) -> DynArray[uint256, 2]:
    ys: DynArray[uint256, 2] = xs
    ys = [ys[1], ys[0]]
    return ys
    """
    c = get_contract(code)
    assert c.bug([1, 2]) == [2, 1]


def test_assign_rhs_lhs_overlap_struct(get_contract):
    # GH issue 2418, generalize to structs
    code = """
struct Point:
    x: uint256
    y: uint256

@external
def bug(p: Point) -> Point:
    t: Point = p
    t = Point(x=t.y, y=t.x)
    return t
    """
    c = get_contract(code)
    assert c.bug((1, 2)) == (2, 1)


mload_merge_codes = [
    (
        """
@external
def foo() -> uint256[4]:
    # copy "backwards"
    xs: uint256[4] = [1, 2, 3, 4]

# dst < src
    xs[0] = xs[1]
    xs[1] = xs[2]
    xs[2] = xs[3]

    return xs
    """,
        [2, 3, 4, 4],
    ),
    (
        """
@external
def foo() -> uint256[4]:
    # copy "forwards"
    xs: uint256[4] = [1, 2, 3, 4]

# src < dst
    xs[1] = xs[0]
    xs[2] = xs[1]
    xs[3] = xs[2]

    return xs
    """,
        [1, 1, 1, 1],
    ),
    (
        """
@external
def foo() -> uint256[5]:
    # partial "forward" copy
    xs: uint256[5] = [1, 2, 3, 4, 5]

# src < dst
    xs[2] = xs[0]
    xs[3] = xs[1]
    xs[4] = xs[2]

    return xs
    """,
        [1, 2, 1, 2, 1],
    ),
]


# functional test that mload merging does not occur when source and dest
# buffers overlap. (note: mload merging only applies after cancun)
@pytest.mark.parametrize("code,expected_result", mload_merge_codes)
def test_mcopy_overlap(get_contract, code, expected_result):
    c = get_contract(code)
    assert c.foo() == expected_result



# -*- coding: utf-8 -*-

import os
import re
import subprocess

from setuptools import setup

extras_require = {
    "test": [
        "pytest>=8.0,<9.0",
        "pytest-cov>=4.1,<5.0",
        "pytest-instafail>=0.4,<1.0",
        "pytest-xdist>=3.0,<3.4",
        "pytest-split>=0.7.0,<1.0",
        "eth_abi>=5.0.0,<6.0.0",
        "py-evm>=0.10.1b1,<0.11",
        "lark==1.1.9",
        "hypothesis[lark]>=6.0,<7.0",
        "eth-stdlib==0.2.7",
        "eth-account==0.12.2",
        "setuptools",
        "hexbytes>=1.2",
        "pyrevm>=0.3.2",
    ],
    "lint": [
        "black==23.12.0",
        "flake8==6.1.0",
        "flake8-bugbear==23.12.2",
        "flake8-use-fstring==1.4",
        "isort==5.13.2",
        "mypy==1.5",
    ],
    "dev": ["ipython", "pre-commit", "pyinstaller", "twine"],
}

extras_require["dev"] = extras_require["dev"] + extras_require["test"] + extras_require["lint"]

with open("README.md", "r") as f:
    long_description = f.read()


# strip local version
def _local_version(version):
    return ""


def _global_version(version):
    from setuptools_scm.version import guess_next_dev_version

    # strip `.devN` suffix since it is not semver compatible
    # minor regex hack to avoid messing too much with setuptools-scm internals
    version_str = guess_next_dev_version(version)
    return re.sub(r"\.dev\d+", "", version_str)


hash_file_rel_path = os.path.join("vyper", "vyper_git_commithash.txt")
hashfile = os.path.relpath(hash_file_rel_path)

# there is no way in setuptools-scm to get metadata besides the package
# version into version.py. (and we need that version to be PEP440 compliant
# in order to get it into pypi). so, add the commit hash to the package
# separately, in order so that we can add it to `vyper --version`.
try:
    commithash = subprocess.check_output("git rev-parse --short HEAD".split())
    commithash_str = commithash.decode("utf-8").strip()
    with open(hashfile, "w") as fh:
        fh.write(commithash_str)
except subprocess.CalledProcessError:
    pass


setup(
    name="vyper",
    use_scm_version={
        "local_scheme": _local_version,
        "version_scheme": _global_version,
        "write_to": "vyper/version.py",
    },
    description="Vyper: the Pythonic Programming Language for the EVM",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Vyper Team",
    author_email="",
    url="https://github.com/vyperlang/vyper",
    license="Apache License 2.0",
    keywords="ethereum evm smart contract language",
    include_package_data=True,
    packages=["vyper"],
    python_requires=">=3.10,<4",
    py_modules=["vyper"],
    install_requires=[
        "cbor2>=5.4.6,<6",
        "asttokens>=2.0.5,<3",
        "pycryptodome>=3.5.1,<4",
        "packaging>=23.1,<24",
        "importlib-metadata",
        "wheel",
    ],
    setup_requires=["pytest-runner", "setuptools_scm>=7.1.0,<8.0.0"],
    tests_require=extras_require["test"],
    extras_require=extras_require,
    entry_points={
        "console_scripts": [
            "vyper=vyper.cli.vyper_compile:_parse_cli_args",
            "fang=vyper.cli.vyper_ir:_parse_cli_args",
            "vyper-json=vyper.cli.vyper_json:_parse_cli_args",
        ]
    },
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    package_data={"vyper.ast": ["grammar.lark"]},
    data_files=[("", [hash_file_rel_path])],
)



import pytest

from vyper import compile_code
from vyper.exceptions import TypeMismatch

good_list = [
    # basic test
    """
@external
def foo(a: uint256, b: uint256) -> uint256:
    return a if a > b else b
    """,
    """
@external
def foo():
    a: bool = (True if True else True) or True
    """,
    # different locations:
    """
b: uint256

@external
def foo(x: uint256) -> uint256:
    return x if x > self.b else self.b
    """,
    # different kinds of test exprs
    """
@external
def foo(x: uint256, t: bool) -> uint256:
    return x if t else 1
    """,
    """
@external
def foo(x: uint256) -> uint256:
    return x if True else 1
    """,
    """
@external
def foo(x: uint256) -> uint256:
    return x if False else 1
    """,
    # more complex types
    """
@external
def foo(t: bool) -> DynArray[uint256, 1]:
    return [2] if t else [1]
    """,
    # TODO: get this working, depends #3377
    #    """
    # @external
    # def foo(t: bool) -> DynArray[uint256, 1]:
    #    return [] if t else [1]
    #    """,
    """
@external
def foo(t: bool) -> (uint256, uint256):
    a: uint256 = 0
    b: uint256 = 1
    return (a, b) if t else (b, a)
    """,
]


@pytest.mark.parametrize("code", good_list)
def test_ternary_good(code):
    assert compile_code(code) is not None


fail_list = [
    (  # bad test type
        """
@external
def foo() -> uint256:
    return 1 if 1 else 2
    """,
        TypeMismatch,
    ),
    (  # bad test type: constant
        """
TEST: constant(uint256) = 1
@external
def foo() -> uint256:
    return 1 if TEST else 2
    """,
        TypeMismatch,
    ),
    (  # bad test type: variable
        """
TEST: constant(uint256) = 1
@external
def foo(t: uint256) -> uint256:
    return 1 if t else 2
    """,
        TypeMismatch,
    ),
    (  # mismatched body and orelse: literal
        """
@external
def foo() -> uint256:
    return 1 if True else 2.0
    """,
        TypeMismatch,
    ),
    (  # mismatched body and orelse: literal and known type
        """
T: constant(uint256) = 1
@external
def foo() -> uint256:
    return T if True else 2.0
    """,
        TypeMismatch,
    ),
    (  # mismatched body and orelse: both variable
        """
@external
def foo(x: uint256, y: uint8) -> uint256:
    return x if True else y
    """,
        TypeMismatch,
    ),
    (  # mismatched tuple types
        """
@external
def foo(a: uint256, b: uint256, c: uint256) -> (uint256, uint256):
    return (a, b) if True else (a, b, c)
    """,
        TypeMismatch,
    ),
    (  # mismatched tuple types - other direction
        """
@external
def foo(a: uint256, b: uint256, c: uint256) -> (uint256, uint256):
    return (a, b, c) if True else (a, b)
    """,
        TypeMismatch,
    ),
]


@pytest.mark.parametrize("code,exc", fail_list)
def test_functions_call_fail(code, exc):
    with pytest.raises(exc):
        compile_code(code)



import pytest

from vyper import compiler
from vyper.exceptions import NamespaceCollision

fail_list = [
    """
@external
def foo(int128: int128):
    pass
    """,
    """
@external
def foo():
    x: int128 = 12
@external
def foo():
    y: int128 = 12
    """,
    """
foo: int128

@external
def foo():
    pass
    """,
    """
x: int128
x: int128
    """,
    """
@external
def foo():
    x: int128 = 0
    x: int128 = 0
    """,
    """
@external
def foo():
    msg: bool = True
    """,
    """
int128: Bytes[3]
    """,
]


@pytest.mark.parametrize("bad_code", fail_list)
def test_insufficient_arguments(bad_code):
    with pytest.raises(NamespaceCollision):
        compiler.compile_code(bad_code)


pass_list = [
    """
x: int128

@external
def foo(x: int128): pass
    """,
    """
x: int128

@external
def foo():
    x: int128 = 1234
    """,
]


@pytest.mark.parametrize("code", pass_list)
def test_valid(code):
    compiler.compile_code(code)



def test_block_number(get_contract, env):
    block_number_code = """
@external
def block_number() -> uint256:
    return block.number
    """
    c = get_contract(block_number_code)

    assert c.block_number() == 1
    env.block_number += 1
    assert c.block_number() == 2



from eth_utils import to_int


def test_log_dynamic_static_combo(get_logs, get_contract):
    code = """
event TestLog:
    testData1: bytes32
    testData2: Bytes[60]
    testData3: Bytes[8]

@internal
@view
def to_bytes(_value: uint256) -> Bytes[8]:
    return slice(concat(b"", convert(_value, bytes32)), 24, 8)

@internal
@view
def to_bytes32(_value: uint256) -> bytes32:
    return convert(_value, bytes32)

@external
def test_func(_value: uint256):
    data2: Bytes[60] = concat(self.to_bytes32(_value),self.to_bytes(_value),b"testing")
    log TestLog(self.to_bytes32(_value), data2, self.to_bytes(_value))

    loggedValue: bytes32 = self.to_bytes32(_value)
    loggedValue2: Bytes[8] = self.to_bytes(_value)
    log TestLog(loggedValue, data2, loggedValue2)
    """

    c = get_contract(code)

    c.test_func(123)
    log, log2 = get_logs(c, "TestLog")

    assert to_int(log.args.testData1) == 123
    assert to_int(log.args.testData2[:32]) == 123
    assert log.args.testData2[-7:] == b"testing"
    assert log.args.testData2[32:] == b"\x00\x00\x00\x00\x00\x00\x00{testing"
    assert log.args.testData3 == b"\x00\x00\x00\x00\x00\x00\x00{"
    assert to_int(log.args.testData3) == 123

    assert log.args == log2.args


def test_log_dynamic_static_combo2(get_logs, get_contract):
    code = """
event TestLog:
    testData1: bytes32
    testData2: Bytes[133]
    testData3: String[8]

@internal
@view
def to_bytes(_value: uint256) -> Bytes[8]:
    return slice(concat(b"", convert(_value, bytes32)), 24, 8)

@internal
@view
def to_bytes32(_value: uint256) -> bytes32:
    return convert(_value, bytes32)

@external
def test_func(_value: uint256,input: Bytes[133]):

    data2: Bytes[200] = b"hello world"

    # log TestLog(self.to_bytes32(_value),input,self.to_bytes(_value))
    log TestLog(self.to_bytes32(_value),input,"bababa")
    """

    c = get_contract(code)

    # assert c.test_func(2**255, b'x' * 129) == b'x' * 129
    c.test_func(1234444, b"x" * 129)
    logs = get_logs(c, "TestLog")

    print(logs[0].args)

    assert to_int(logs[0].args.testData1) == 1234444
    assert logs[0].args.testData3 == "bababa"
    assert logs[0].args.testData2 == b"x" * 129


def test_log_single_function_call(get_logs, get_contract):
    code = """
event TestLog:
    testData1: bytes32
    testData2: Bytes[133]

@internal
@view
def to_bytes32(_value: uint256) -> bytes32:
    return convert(_value, bytes32)

@external
def test_func(_value: uint256,input: Bytes[133]):

    data2: Bytes[200] = b"hello world"

    # log will be malformed
    # log TestLog(self.to_bytes32(_value),input,self.to_bytes(_value))
    log TestLog(self.to_bytes32(_value), input)
    """

    c = get_contract(code)

    c.test_func(1234444, b"x" * 129)
    (log,) = get_logs(c, "TestLog")

    assert to_int(log.args.testData1) == 1234444
    assert log.args.testData2 == b"x" * 129


def test_original_problem_function(get_logs, get_contract):
    # See #1205 for further details, this is kept as test case as it introduces very specific
    # edge cases to the ABI encoder when logging.
    code = """
event TestLog:
    testData1: bytes32
    testData2: Bytes[2064]
    testData3: Bytes[8]

@internal
@view
def to_bytes(_value: uint256) -> Bytes[8]:
    return slice(concat(b"", convert(_value, bytes32)), 24, 8)

@internal
@view
def to_bytes32(_value: uint256) -> bytes32:
    return convert(_value, bytes32)

@external
def test_func(_value: uint256,input: Bytes[2048]):

    data2: Bytes[2064] = concat(self.to_bytes(_value),self.to_bytes(_value),input)

    # log will be malformed
    log TestLog(self.to_bytes32(_value), data2, self.to_bytes(_value))

    loggedValue: Bytes[8] = self.to_bytes(_value)

    # log will be normal
    log TestLog(self.to_bytes32(_value),data2,loggedValue)
    """

    c = get_contract(code)

    c.test_func(333, b"x" * 132)
    log1, log2 = get_logs(c, "TestLog")

    print(log1.args)

    assert to_int(log1.args.testData1) == 333
    assert to_int(log1.args.testData2[0:8]) == 333
    assert to_int(log1.args.testData2[8:16]) == 333
    assert log1.args.testData2[16:] == b"x" * 132
    assert to_int(log1.args.testData3) == 333

    assert log1.args == log2.args



import pytest

from vyper import compile_code

valid_list = [
    """
BAR: constant(decimal) = 2.5
FOO: constant(int256) = ceil(BAR)

@external
def foo():
    a: int256 = FOO
    """
]


@pytest.mark.parametrize("code", valid_list)
def test_ceil_good(code):
    assert compile_code(code) is not None



import pytest

from vyper import compile_code

valid_list = [
    """
FOO: constant(uint256) = 3
BAR: constant(String[78]) = uint2str(FOO)

@external
def foo():
    a: String[78] = BAR
    """
]


@pytest.mark.parametrize("code", valid_list)
def test_addmulmod_pass(code):
    assert compile_code(code) is not None



def test_arbitration_code(env, get_contract, tx_failed):
    arbitration_code = """
buyer: address
seller: address
arbitrator: address

@external
def setup(_seller: address, _arbitrator: address):
    if self.buyer == empty(address):
        self.buyer = msg.sender
        self.seller = _seller
        self.arbitrator = _arbitrator

@external
def finalize():
    assert msg.sender == self.buyer or msg.sender == self.arbitrator
    send(self.seller, self.balance)

@external
def refund():
    assert msg.sender == self.seller or msg.sender == self.arbitrator
    send(self.buyer, self.balance)

    """
    a0, a1, a2 = env.accounts[:3]
    env.set_balance(a0, 1)
    c = get_contract(arbitration_code, value=1)
    c.setup(a1, a2)
    with tx_failed():
        c.finalize(sender=a1)
    c.finalize()

    print("Passed escrow test")


def test_arbitration_code_with_init(env, tx_failed, get_contract):
    arbitration_code_with_init = """
buyer: address
seller: address
arbitrator: address

@deploy
@payable
def __init__(_seller: address, _arbitrator: address):
    if self.buyer == empty(address):
        self.buyer = msg.sender
        self.seller = _seller
        self.arbitrator = _arbitrator

@external
def finalize():
    assert msg.sender == self.buyer or msg.sender == self.arbitrator
    send(self.seller, self.balance)

@external
def refund():
    assert msg.sender == self.seller or msg.sender == self.arbitrator
    send(self.buyer, self.balance)
    """
    a0, a1, a2 = env.accounts[:3]
    env.set_balance(env.deployer, 1)
    c = get_contract(arbitration_code_with_init, *[a1, a2], value=1)
    with tx_failed():
        c.finalize(sender=a1)
    c.finalize(sender=a0)

    print("Passed escrow test with initializer")



from dataclasses import dataclass
from enum import Enum
from functools import reduce
from typing import Union

from vyper.exceptions import CompilerPanic, StaticAssertionException
from vyper.utils import OrderedSet
from vyper.venom.analysis.analysis import IRAnalysesCache
from vyper.venom.analysis.cfg import CFGAnalysis
from vyper.venom.analysis.dominators import DominatorTreeAnalysis
from vyper.venom.basicblock import (
    IRBasicBlock,
    IRInstruction,
    IRLabel,
    IRLiteral,
    IROperand,
    IRVariable,
)
from vyper.venom.function import IRFunction
from vyper.venom.passes.base_pass import IRPass
from vyper.venom.passes.sccp.eval import ARITHMETIC_OPS


class LatticeEnum(Enum):
    TOP = 1
    BOTTOM = 2


@dataclass
class SSAWorkListItem:
    inst: IRInstruction


@dataclass
class FlowWorkItem:
    start: IRBasicBlock
    end: IRBasicBlock


WorkListItem = Union[FlowWorkItem, SSAWorkListItem]
LatticeItem = Union[LatticeEnum, IRLiteral]
Lattice = dict[IRVariable, LatticeItem]


class SCCP(IRPass):
    """
    This class implements the Sparse Conditional Constant Propagation
    algorithm by Wegman and Zadeck. It is a forward dataflow analysis
    that propagates constant values through the IR graph. It is used
    to optimize the IR by removing dead code and replacing variables
    with their constant values.
    """

    fn: IRFunction
    dom: DominatorTreeAnalysis
    uses: dict[IRVariable, OrderedSet[IRInstruction]]
    lattice: Lattice
    work_list: list[WorkListItem]
    cfg_dirty: bool
    cfg_in_exec: dict[IRBasicBlock, OrderedSet[IRBasicBlock]]

    def __init__(self, analyses_cache: IRAnalysesCache, function: IRFunction):
        super().__init__(analyses_cache, function)
        self.lattice = {}
        self.work_list: list[WorkListItem] = []
        self.cfg_dirty = False

    def run_pass(self):
        self.fn = self.function
        self.dom = self.analyses_cache.request_analysis(DominatorTreeAnalysis)
        self._compute_uses()
        self._calculate_sccp(self.fn.entry)
        self._propagate_constants()

        # self._propagate_variables()

        self.analyses_cache.invalidate_analysis(CFGAnalysis)

    def _calculate_sccp(self, entry: IRBasicBlock):
        """
        This method is the main entry point for the SCCP algorithm. It
        initializes the work list and the lattice and then iterates over
        the work list until it is empty. It then visits each basic block
        in the CFG and processes the instructions in the block.

        This method does not update the IR, it only updates the lattice
        and the work list. The `_propagate_constants()` method is responsible
        for updating the IR with the constant values.
        """
        self.cfg_in_exec = {bb: OrderedSet() for bb in self.fn.get_basic_blocks()}

        dummy = IRBasicBlock(IRLabel("__dummy_start"), self.fn)
        self.work_list.append(FlowWorkItem(dummy, entry))

        # Initialize the lattice with TOP values for all variables
        for v in self.uses.keys():
            self.lattice[v] = LatticeEnum.TOP

        # Iterate over the work list until it is empty
        # Items in the work list can be either FlowWorkItem or SSAWorkListItem
        while len(self.work_list) > 0:
            work_item = self.work_list.pop()
            if isinstance(work_item, FlowWorkItem):
                self._handle_flow_work_item(work_item)
            elif isinstance(work_item, SSAWorkListItem):
                self._handle_SSA_work_item(work_item)
            else:
                raise CompilerPanic("Invalid work item type")

    def _handle_flow_work_item(self, work_item: FlowWorkItem):
        """
        This method handles a FlowWorkItem.
        """
        start = work_item.start
        end = work_item.end
        if start in self.cfg_in_exec[end]:
            return
        self.cfg_in_exec[end].add(start)

        for inst in end.instructions:
            if inst.opcode == "phi":
                self._visit_phi(inst)
            else:
                # Stop at the first non-phi instruction
                # as phis are only valid at the beginning of a block
                break

        if len(self.cfg_in_exec[end]) == 1:
            for inst in end.instructions:
                if inst.opcode == "phi":
                    continue
                self._visit_expr(inst)

        if len(end.cfg_out) == 1:
            self.work_list.append(FlowWorkItem(end, end.cfg_out.first()))

    def _handle_SSA_work_item(self, work_item: SSAWorkListItem):
        """
        This method handles a SSAWorkListItem.
        """
        if work_item.inst.opcode == "phi":
            self._visit_phi(work_item.inst)
        elif len(self.cfg_in_exec[work_item.inst.parent]) > 0:
            self._visit_expr(work_item.inst)

    def _lookup_from_lattice(self, op: IROperand) -> LatticeItem:
        assert isinstance(op, IRVariable), "Can't get lattice for non-variable"
        lat = self.lattice[op]
        assert lat is not None, f"Got undefined var {op}"
        return lat

    def _set_lattice(self, op: IROperand, value: LatticeItem):
        assert isinstance(op, IRVariable), "Can't set lattice for non-variable"
        self.lattice[op] = value

    def _eval_from_lattice(self, op: IROperand) -> IRLiteral | LatticeEnum:
        if isinstance(op, IRLiteral):
            return op

        return self._lookup_from_lattice(op)

    def _visit_phi(self, inst: IRInstruction):
        assert inst.opcode == "phi", "Can't visit non phi instruction"
        in_vars: list[LatticeItem] = []
        for bb_label, var in inst.phi_operands:
            bb = self.fn.get_basic_block(bb_label.name)
            if bb not in self.cfg_in_exec[inst.parent]:
                continue
            in_vars.append(self._lookup_from_lattice(var))
        value = reduce(_meet, in_vars, LatticeEnum.TOP)  # type: ignore

        if inst.output not in self.lattice:
            return

        if value != self._lookup_from_lattice(inst.output):
            self._set_lattice(inst.output, value)
            self._add_ssa_work_items(inst)

    def _visit_expr(self, inst: IRInstruction):
        opcode = inst.opcode
        if opcode in ["store", "alloca"]:
            assert inst.output is not None, "Got store/alloca without output"
            out = self._eval_from_lattice(inst.operands[0])
            self._set_lattice(inst.output, out)
            self._add_ssa_work_items(inst)
        elif opcode == "jmp":
            target = self.fn.get_basic_block(inst.operands[0].value)
            self.work_list.append(FlowWorkItem(inst.parent, target))
        elif opcode == "jnz":
            lat = self._eval_from_lattice(inst.operands[0])

            assert lat != LatticeEnum.TOP, f"Got undefined var at jmp at {inst.parent}"
            if lat == LatticeEnum.BOTTOM:
                for out_bb in inst.parent.cfg_out:
                    self.work_list.append(FlowWorkItem(inst.parent, out_bb))
            else:
                if _meet(lat, IRLiteral(0)) == LatticeEnum.BOTTOM:
                    target = self.fn.get_basic_block(inst.operands[1].name)
                    self.work_list.append(FlowWorkItem(inst.parent, target))
                if _meet(lat, IRLiteral(1)) == LatticeEnum.BOTTOM:
                    target = self.fn.get_basic_block(inst.operands[2].name)
                    self.work_list.append(FlowWorkItem(inst.parent, target))
        elif opcode == "djmp":
            lat = self._eval_from_lattice(inst.operands[0])
            assert lat != LatticeEnum.TOP, f"Got undefined var at jmp at {inst.parent}"
            if lat == LatticeEnum.BOTTOM:
                for op in inst.operands[1:]:
                    target = self.fn.get_basic_block(op.name)
                    self.work_list.append(FlowWorkItem(inst.parent, target))
            elif isinstance(lat, IRLiteral):
                raise CompilerPanic("Unimplemented djmp with literal")

        elif opcode in ["param", "calldataload"]:
            self.lattice[inst.output] = LatticeEnum.BOTTOM  # type: ignore
            self._add_ssa_work_items(inst)
        elif opcode == "mload":
            self.lattice[inst.output] = LatticeEnum.BOTTOM  # type: ignore
        elif opcode in ARITHMETIC_OPS:
            self._eval(inst)
        else:
            if inst.output is not None:
                self._set_lattice(inst.output, LatticeEnum.BOTTOM)

    def _eval(self, inst) -> LatticeItem:
        """
        This method evaluates an arithmetic operation and returns the result.
        At the same time it updates the lattice with the result and adds the
        instruction to the SSA work list if the knowledge about the variable
        changed.
        """
        opcode = inst.opcode

        ops = []
        for op in inst.operands:
            if isinstance(op, IRVariable):
                ops.append(self.lattice[op])
            elif isinstance(op, IRLabel):
                return LatticeEnum.BOTTOM
            else:
                ops.append(op)

        ret = None
        if LatticeEnum.BOTTOM in ops:
            ret = LatticeEnum.BOTTOM
        else:
            if opcode in ARITHMETIC_OPS:
                fn = ARITHMETIC_OPS[opcode]
                ret = IRLiteral(fn(ops))  # type: ignore
            elif len(ops) > 0:
                ret = ops[0]  # type: ignore
            else:
                raise CompilerPanic("Bad constant evaluation")

        old_val = self.lattice.get(inst.output, LatticeEnum.TOP)
        if old_val != ret:
            self.lattice[inst.output] = ret  # type: ignore
            self._add_ssa_work_items(inst)

        return ret  # type: ignore

    def _add_ssa_work_items(self, inst: IRInstruction):
        for target_inst in self._get_uses(inst.output):  # type: ignore
            self.work_list.append(SSAWorkListItem(target_inst))

    def _compute_uses(self):
        """
        This method computes the uses for each variable in the IR.
        It iterates over the dominator tree and collects all the
        instructions that use each variable.
        """
        self.uses = {}
        for bb in self.dom.dfs_walk:
            for var, insts in bb.get_uses().items():
                self._get_uses(var).update(insts)

    def _get_uses(self, var: IRVariable):
        if var not in self.uses:
            self.uses[var] = OrderedSet()
        return self.uses[var]

    def _propagate_constants(self):
        """
        This method iterates over the IR and replaces constant values
        with their actual values. It also replaces conditional jumps
        with unconditional jumps if the condition is a constant value.
        """
        for bb in self.dom.dfs_walk:
            for inst in bb.instructions:
                self._replace_constants(inst)

    def _replace_constants(self, inst: IRInstruction):
        """
        This method replaces constant values in the instruction with
        their actual values. It also updates the instruction opcode in
        case of jumps and asserts as needed.
        """
        if inst.opcode == "jnz":
            lat = self._eval_from_lattice(inst.operands[0])

            if isinstance(lat, IRLiteral):
                if lat.value == 0:
                    target = inst.operands[2]
                else:
                    target = inst.operands[1]
                inst.opcode = "jmp"
                inst.operands = [target]
                self.cfg_dirty = True

        elif inst.opcode in ("assert", "assert_unreachable"):
            lat = self._eval_from_lattice(inst.operands[0])

            if isinstance(lat, IRLiteral):
                if lat.value > 0:
                    inst.opcode = "nop"
                else:
                    raise StaticAssertionException(
                        f"assertion found to fail at compile time ({inst.error_msg}).",
                        inst.get_ast_source(),
                    )

                inst.operands = []

        elif inst.opcode == "phi":
            return

        for i, op in enumerate(inst.operands):
            if isinstance(op, IRVariable):
                lat = self.lattice[op]
                if isinstance(lat, IRLiteral):
                    inst.operands[i] = lat


def _meet(x: LatticeItem, y: LatticeItem) -> LatticeItem:
    if x == LatticeEnum.TOP:
        return y
    if y == LatticeEnum.TOP or x == y:
        return x
    return LatticeEnum.BOTTOM



from typing import Dict, Optional

from vyper.compiler.settings import get_global_settings
from vyper.exceptions import CompilerPanic
from vyper.typing import OpcodeGasCost, OpcodeMap, OpcodeRulesetMap, OpcodeRulesetValue, OpcodeValue

# EVM version rules work as follows:
# 1. Fork rules go from oldest (lowest value) to newest (highest value).
# 2. Fork versions aren't actually tied to anything. They are not a part of our
#    official API. *DO NOT USE THE VALUES FOR ANYTHING IMPORTANT* besides versioning.
# 3. Per VIP-3365, we support mainnet fork choice rules up to 3 years old
#    (and may optionally have forward support for experimental/unreleased
#    fork choice rules)
_evm_versions = ("london", "paris", "shanghai", "cancun")
EVM_VERSIONS: dict[str, int] = dict((v, i) for i, v in enumerate(_evm_versions))

DEFAULT_EVM_VERSION = "shanghai"


# opcode as hex value
# number of values removed from stack
# number of values added to stack
# gas cost (london, paris, shanghai, cancun)
OPCODES: OpcodeMap = {
    "STOP": (0x00, 0, 0, 0),
    "ADD": (0x01, 2, 1, 3),
    "MUL": (0x02, 2, 1, 5),
    "SUB": (0x03, 2, 1, 3),
    "DIV": (0x04, 2, 1, 5),
    "SDIV": (0x05, 2, 1, 5),
    "MOD": (0x06, 2, 1, 5),
    "SMOD": (0x07, 2, 1, 5),
    "ADDMOD": (0x08, 3, 1, 8),
    "MULMOD": (0x09, 3, 1, 8),
    "EXP": (0x0A, 2, 1, 10),
    "SIGNEXTEND": (0x0B, 2, 1, 5),
    "LT": (0x10, 2, 1, 3),
    "GT": (0x11, 2, 1, 3),
    "SLT": (0x12, 2, 1, 3),
    "SGT": (0x13, 2, 1, 3),
    "EQ": (0x14, 2, 1, 3),
    "ISZERO": (0x15, 1, 1, 3),
    "AND": (0x16, 2, 1, 3),
    "OR": (0x17, 2, 1, 3),
    "XOR": (0x18, 2, 1, 3),
    "NOT": (0x19, 1, 1, 3),
    "BYTE": (0x1A, 2, 1, 3),
    "SHL": (0x1B, 2, 1, 3),
    "SHR": (0x1C, 2, 1, 3),
    "SAR": (0x1D, 2, 1, 3),
    "SHA3": (0x20, 2, 1, 30),
    "ADDRESS": (0x30, 0, 1, 2),
    "BALANCE": (0x31, 1, 1, 700),
    "ORIGIN": (0x32, 0, 1, 2),
    "CALLER": (0x33, 0, 1, 2),
    "CALLVALUE": (0x34, 0, 1, 2),
    "CALLDATALOAD": (0x35, 1, 1, 3),
    "CALLDATASIZE": (0x36, 0, 1, 2),
    "CALLDATACOPY": (0x37, 3, 0, 3),
    "CODESIZE": (0x38, 0, 1, 2),
    "CODECOPY": (0x39, 3, 0, 3),
    "GASPRICE": (0x3A, 0, 1, 2),
    "EXTCODESIZE": (0x3B, 1, 1, 2600),
    "EXTCODECOPY": (0x3C, 4, 0, 2600),
    "RETURNDATASIZE": (0x3D, 0, 1, 2),
    "RETURNDATACOPY": (0x3E, 3, 0, 3),
    "EXTCODEHASH": (0x3F, 1, 1, 2600),
    "BLOCKHASH": (0x40, 1, 1, 20),
    "COINBASE": (0x41, 0, 1, 2),
    "TIMESTAMP": (0x42, 0, 1, 2),
    "NUMBER": (0x43, 0, 1, 2),
    "DIFFICULTY": (0x44, 0, 1, 2),
    "PREVRANDAO": (0x44, 0, 1, 2),
    "GASLIMIT": (0x45, 0, 1, 2),
    "CHAINID": (0x46, 0, 1, 2),
    "SELFBALANCE": (0x47, 0, 1, 5),
    "BASEFEE": (0x48, 0, 1, 2),
    "BLOBHASH": (0x49, 1, 1, (None, None, None, 3)),
    "BLOBBASEFEE": (0x4A, 0, 1, (None, None, None, 2)),
    "POP": (0x50, 1, 0, 2),
    "MLOAD": (0x51, 1, 1, 3),
    "MSTORE": (0x52, 2, 0, 3),
    "MSTORE8": (0x53, 2, 0, 3),
    "SLOAD": (0x54, 1, 1, 2100),
    "SSTORE": (0x55, 2, 0, 20000),
    "JUMP": (0x56, 1, 0, 8),
    "JUMPI": (0x57, 2, 0, 10),
    "PC": (0x58, 0, 1, 2),
    "MSIZE": (0x59, 0, 1, 2),
    "GAS": (0x5A, 0, 1, 2),
    "JUMPDEST": (0x5B, 0, 0, 1),
    "MCOPY": (0x5E, 3, 0, (None, None, None, 3)),
    "PUSH0": (0x5F, 0, 1, 2),
    "PUSH1": (0x60, 0, 1, 3),
    "PUSH2": (0x61, 0, 1, 3),
    "PUSH3": (0x62, 0, 1, 3),
    "PUSH4": (0x63, 0, 1, 3),
    "PUSH5": (0x64, 0, 1, 3),
    "PUSH6": (0x65, 0, 1, 3),
    "PUSH7": (0x66, 0, 1, 3),
    "PUSH8": (0x67, 0, 1, 3),
    "PUSH9": (0x68, 0, 1, 3),
    "PUSH10": (0x69, 0, 1, 3),
    "PUSH11": (0x6A, 0, 1, 3),
    "PUSH12": (0x6B, 0, 1, 3),
    "PUSH13": (0x6C, 0, 1, 3),
    "PUSH14": (0x6D, 0, 1, 3),
    "PUSH15": (0x6E, 0, 1, 3),
    "PUSH16": (0x6F, 0, 1, 3),
    "PUSH17": (0x70, 0, 1, 3),
    "PUSH18": (0x71, 0, 1, 3),
    "PUSH19": (0x72, 0, 1, 3),
    "PUSH20": (0x73, 0, 1, 3),
    "PUSH21": (0x74, 0, 1, 3),
    "PUSH22": (0x75, 0, 1, 3),
    "PUSH23": (0x76, 0, 1, 3),
    "PUSH24": (0x77, 0, 1, 3),
    "PUSH25": (0x78, 0, 1, 3),
    "PUSH26": (0x79, 0, 1, 3),
    "PUSH27": (0x7A, 0, 1, 3),
    "PUSH28": (0x7B, 0, 1, 3),
    "PUSH29": (0x7C, 0, 1, 3),
    "PUSH30": (0x7D, 0, 1, 3),
    "PUSH31": (0x7E, 0, 1, 3),
    "PUSH32": (0x7F, 0, 1, 3),
    "DUP1": (0x80, 1, 2, 3),
    "DUP2": (0x81, 1, 2, 3),
    "DUP3": (0x82, 1, 2, 3),
    "DUP4": (0x83, 1, 2, 3),
    "DUP5": (0x84, 1, 2, 3),
    "DUP6": (0x85, 1, 2, 3),
    "DUP7": (0x86, 1, 2, 3),
    "DUP8": (0x87, 1, 2, 3),
    "DUP9": (0x88, 1, 2, 3),
    "DUP10": (0x89, 1, 2, 3),
    "DUP11": (0x8A, 1, 2, 3),
    "DUP12": (0x8B, 1, 2, 3),
    "DUP13": (0x8C, 1, 2, 3),
    "DUP14": (0x8D, 1, 2, 3),
    "DUP15": (0x8E, 1, 2, 3),
    "DUP16": (0x8F, 1, 2, 3),
    "SWAP1": (0x90, 2, 2, 3),
    "SWAP2": (0x91, 2, 2, 3),
    "SWAP3": (0x92, 2, 2, 3),
    "SWAP4": (0x93, 2, 2, 3),
    "SWAP5": (0x94, 2, 2, 3),
    "SWAP6": (0x95, 2, 2, 3),
    "SWAP7": (0x96, 2, 2, 3),
    "SWAP8": (0x97, 2, 2, 3),
    "SWAP9": (0x98, 2, 2, 3),
    "SWAP10": (0x99, 2, 2, 3),
    "SWAP11": (0x9A, 2, 2, 3),
    "SWAP12": (0x9B, 2, 2, 3),
    "SWAP13": (0x9C, 2, 2, 3),
    "SWAP14": (0x9D, 2, 2, 3),
    "SWAP15": (0x9E, 2, 2, 3),
    "SWAP16": (0x9F, 2, 2, 3),
    "LOG0": (0xA0, 2, 0, 375),
    "LOG1": (0xA1, 3, 0, 750),
    "LOG2": (0xA2, 4, 0, 1125),
    "LOG3": (0xA3, 5, 0, 1500),
    "LOG4": (0xA4, 6, 0, 1875),
    "CREATE": (0xF0, 3, 1, 32000),
    "CALL": (0xF1, 7, 1, 2100),
    "CALLCODE": (0xF2, 7, 1, 2100),
    "RETURN": (0xF3, 2, 0, 0),
    "DELEGATECALL": (0xF4, 6, 1, 2100),
    "CREATE2": (0xF5, 4, 1, 32000),
    "SELFDESTRUCT": (0xFF, 1, 0, 25000),
    "STATICCALL": (0xFA, 6, 1, 2100),
    "REVERT": (0xFD, 2, 0, 0),
    "INVALID": (0xFE, 0, 0, 0),
    "DEBUG": (0xA5, 1, 0, 0),
    "BREAKPOINT": (0xA6, 0, 0, 0),
    "TLOAD": (0x5C, 1, 1, (None, None, None, 100)),
    "TSTORE": (0x5D, 2, 0, (None, None, None, 100)),
}

PSEUDO_OPCODES: OpcodeMap = {
    "CLAMP": (None, 3, 1, 70),
    "UCLAMPLT": (None, 2, 1, 25),
    "UCLAMPLE": (None, 2, 1, 30),
    "CLAMP_NONZERO": (None, 1, 1, 19),
    "ASSERT": (None, 1, 0, 85),
    "ASSERT_UNREACHABLE": (None, 1, 0, 17),
    "PASS": (None, 0, 0, 0),
    "DUMMY": (None, 0, 1, 0),  # tell IR that no, there really is a stack item here
    "BREAK": (None, 0, 0, 20),
    # cleanup_repeat cleans the stack similar to BREAK but without jumping to exit
    "CLEANUP_REPEAT": (None, 0, 0, 20),
    "CONTINUE": (None, 0, 0, 20),
    "SHA3_32": (None, 1, 1, 72),
    "SHA3_64": (None, 2, 1, 109),
    "SLE": (None, 2, 1, 10),
    "SGE": (None, 2, 1, 10),
    "LE": (None, 2, 1, 10),
    "GE": (None, 2, 1, 10),
    "CEIL32": (None, 1, 1, 20),
    "SET": (None, 2, 0, 20),
    "NE": (None, 2, 1, 6),
    "DEBUGGER": (None, 0, 0, 0),
    "ILOAD": (None, 1, 1, 6),
    "ISTORE": (None, 2, 0, 6),
    "DLOAD": (None, 1, 1, 9),
    "DLOADBYTES": (None, 3, 0, 3),
}

IR_OPCODES: OpcodeMap = {**OPCODES, **PSEUDO_OPCODES}


def _gas(value: OpcodeValue, idx: int) -> Optional[OpcodeRulesetValue]:
    gas: OpcodeGasCost = value[3]
    if isinstance(gas, int):
        return value[:3] + (gas,)
    if len(gas) <= idx:
        return value[:3] + (gas[-1],)
    if gas[idx] is None:
        return None
    return value[:3] + (gas[idx],)


def _mk_version_opcodes(opcodes: OpcodeMap, idx: int) -> OpcodeRulesetMap:
    return dict(
        (k, _gas(v, idx)) for k, v in opcodes.items() if _gas(v, idx) is not None  # type: ignore
    )


_evm_opcodes: Dict[int, OpcodeRulesetMap] = {
    v: _mk_version_opcodes(OPCODES, v) for v in EVM_VERSIONS.values()
}
_ir_opcodes: Dict[int, OpcodeRulesetMap] = {
    v: _mk_version_opcodes(IR_OPCODES, v) for v in EVM_VERSIONS.values()
}


def get_active_evm_version():
    settings = get_global_settings()
    evm_version_str = settings and settings.evm_version or DEFAULT_EVM_VERSION
    return EVM_VERSIONS[evm_version_str]


def get_opcodes() -> OpcodeRulesetMap:
    return _evm_opcodes[get_active_evm_version()]


def get_ir_opcodes() -> OpcodeRulesetMap:
    return _ir_opcodes[get_active_evm_version()]


def version_check(begin: Optional[str] = None, end: Optional[str] = None) -> bool:
    active_evm_version = get_active_evm_version()

    if begin is None and end is None:
        raise CompilerPanic("Either beginning or end fork ruleset must be set.")
    if begin is None:
        begin_idx = min(EVM_VERSIONS.values())
    else:
        begin_idx = EVM_VERSIONS[begin]
    end_idx = max(EVM_VERSIONS.values()) if end is None else EVM_VERSIONS[end]
    return begin_idx <= active_evm_version <= end_idx



from typing import Type

from vyper.venom.function import IRFunction


class IRAnalysis:
    """
    Base class for all Venom IR analyses.
    """

    function: "IRFunction"
    analyses_cache: "IRAnalysesCache"

    def __init__(self, analyses_cache: "IRAnalysesCache", function: IRFunction):
        self.analyses_cache = analyses_cache
        self.function = function

    def analyze(self, *args, **kwargs):
        """
        Override this method to perform the analysis.
        """
        raise NotImplementedError

    def invalidate(self):
        """
        Override this method to respond to an invalidation request, and possibly
        invalidate any other analyses that depend on this one.
        """
        pass


class IRAnalysesCache:
    """
    A cache for IR analyses.
    """

    function: IRFunction
    analyses_cache: dict[Type[IRAnalysis], IRAnalysis]

    def __init__(self, function: IRFunction):
        self.analyses_cache = {}
        self.function = function

    def request_analysis(self, analysis_cls: Type[IRAnalysis], *args, **kwargs):
        """
        Request a specific analysis to be run on the IR. The result is cached and
        returned if the analysis has already been run.
        """
        assert issubclass(analysis_cls, IRAnalysis), f"{analysis_cls} is not an IRAnalysis"
        if analysis_cls in self.analyses_cache:
            return self.analyses_cache[analysis_cls]
        analysis = analysis_cls(self, self.function)
        analysis.analyze(*args, **kwargs)

        self.analyses_cache[analysis_cls] = analysis
        return analysis

    def invalidate_analysis(self, analysis_cls: Type[IRAnalysis]):
        """
        Invalidate a specific analysis. This will remove the analysis from the cache.
        """
        assert issubclass(analysis_cls, IRAnalysis), f"{analysis_cls} is not an IRAnalysis"
        analysis = self.analyses_cache.pop(analysis_cls, None)
        if analysis is not None:
            analysis.invalidate()

    def force_analysis(self, analysis_cls: Type[IRAnalysis], *args, **kwargs):
        """
        Force a specific analysis to be run on the IR even if it has already been run,
        and is cached.
        """
        assert issubclass(analysis_cls, IRAnalysis), f"{analysis_cls} is not an IRAnalysis"
        if analysis_cls in self.analyses_cache:
            self.invalidate_analysis(analysis_cls)
        return self.request_analysis(analysis_cls, *args, **kwargs)



# not an entry point!
# utility functions to handle compiling from a "vyper archive"

import base64
import binascii
import io
import json
import zipfile
from pathlib import PurePath

from vyper.compiler import compile_from_file_input
from vyper.compiler.input_bundle import FileInput, ZipInputBundle
from vyper.compiler.settings import Settings, merge_settings
from vyper.exceptions import BadArchive


class NotZipInput(Exception):
    pass


def compile_from_zip(file_name, output_formats, settings, no_bytecode_metadata):
    with open(file_name, "rb") as f:
        bcontents = f.read()

    try:
        buf = io.BytesIO(bcontents)
        archive = zipfile.ZipFile(buf, mode="r")
    except zipfile.BadZipFile as e1:
        try:
            # `validate=False` - tools like base64 can generate newlines
            # for readability. validate=False does the "correct" thing and
            # simply ignores these
            bcontents = base64.b64decode(bcontents, validate=False)
            buf = io.BytesIO(bcontents)
            archive = zipfile.ZipFile(buf, mode="r")
        except (zipfile.BadZipFile, binascii.Error):
            raise NotZipInput() from e1

    fcontents = archive.read("MANIFEST/compilation_targets").decode("utf-8")
    compilation_targets = fcontents.splitlines()

    if len(compilation_targets) != 1:
        raise BadArchive("Multiple compilation targets not supported!")

    input_bundle = ZipInputBundle(archive)

    mainpath = PurePath(compilation_targets[0])
    file = input_bundle.load_file(mainpath)
    assert isinstance(file, FileInput)  # mypy hint

    settings = settings or Settings()

    archive_settings_txt = archive.read("MANIFEST/settings.json").decode("utf-8")
    archive_settings = Settings.from_dict(json.loads(archive_settings_txt))

    integrity = archive.read("MANIFEST/integrity").decode("utf-8").strip()

    settings = merge_settings(
        settings, archive_settings, lhs_source="command line", rhs_source="archive settings"
    )

    # TODO: validate integrity sum (probably in CompilerData)
    return compile_from_file_input(
        file,
        input_bundle=input_bundle,
        output_formats=output_formats,
        integrity_sum=integrity,
        settings=settings,
        no_bytecode_metadata=no_bytecode_metadata,
    )



import enum
import itertools
import math
from decimal import Decimal

import eth.codecs.abi as abi
import eth.codecs.abi.exceptions
import pytest

from tests.utils import decimal_to_int
from vyper.compiler import compile_code
from vyper.exceptions import InvalidLiteral, InvalidType, TypeMismatch
from vyper.semantics.types import AddressT, BoolT, BytesM_T, BytesT, DecimalT, IntegerT, StringT
from vyper.semantics.types.shortcuts import BYTES20_T, BYTES32_T, UINT, UINT160_T, UINT256_T
from vyper.utils import (
    DECIMAL_DIVISOR,
    checksum_encode,
    int_bounds,
    is_checksum_encoded,
    quantize,
    round_towards_zero,
    unsigned_to_signed,
)

BASE_TYPES = set(IntegerT.all()) | set(BytesM_T.all()) | {DecimalT(), AddressT(), BoolT()}

TEST_TYPES = BASE_TYPES | {BytesT(32)} | {StringT(32)}

# decimal increment, aka smallest decimal > 0
DECIMAL_EPSILON = Decimal(1) / DECIMAL_DIVISOR


def _bits_of_type(typ):
    if isinstance(typ, (IntegerT, DecimalT)):
        return typ.bits
    if isinstance(typ, BoolT):
        return 8
    if isinstance(typ, AddressT):
        return 160
    if isinstance(typ, BytesM_T):
        return typ.m_bits
    if isinstance(typ, BytesT):
        return typ.length * 8

    raise Exception(f"Unknown type {typ}")


def bytes_of_type(typ):
    ret = _bits_of_type(typ)
    assert ret % 8 == 0
    return ret // 8


class _OutOfBounds(Exception):
    """
    A Python-level conversion is out of bounds
    """

    pass


def can_convert(i_typ, o_typ):
    """
    Checks whether conversion from one type to another is valid.
    """
    if i_typ == o_typ:
        return False

    if isinstance(o_typ, BoolT):
        return True
    if isinstance(i_typ, BoolT):
        return not isinstance(o_typ, AddressT)

    if isinstance(i_typ, IntegerT):
        if isinstance(o_typ, BytesM_T):
            return bytes_of_type(i_typ) <= bytes_of_type(o_typ)

        ret = isinstance(o_typ, (IntegerT, DecimalT, BytesM_T, BytesT))
        if not i_typ.is_signed:
            ret |= isinstance(o_typ, AddressT)
        return ret

    if isinstance(i_typ, BytesM_T):
        if isinstance(o_typ, BytesT):
            # bytesN must be of equal or smaller size to the input
            return bytes_of_type(i_typ) <= bytes_of_type(o_typ)

        return isinstance(o_typ, (DecimalT, BytesM_T, IntegerT, AddressT))

    if isinstance(i_typ, BytesT):
        return isinstance(o_typ, (IntegerT, DecimalT, AddressT))

    if isinstance(i_typ, DecimalT):
        if isinstance(o_typ, BytesM_T):
            return bytes_of_type(i_typ) <= bytes_of_type(o_typ)

        return isinstance(o_typ, (IntegerT, BoolT))

    if isinstance(i_typ, AddressT):
        if isinstance(o_typ, BytesM_T):
            return bytes_of_type(i_typ) <= bytes_of_type(o_typ)
        if isinstance(o_typ, IntegerT):
            return not o_typ.is_signed
        return False

    raise AssertionError(f"unreachable {i_typ} {o_typ}")


def uniq(xs):
    return list(set(xs))


def _cases_for_int(typ):
    lo, hi = typ.ast_bounds

    ret = [lo - 1, lo, lo + 1, -1, 0, 1, hi - 1, hi, hi + 1]

    # random cases cause reproducibility issues. TODO fixme
    # NUM_RANDOM_CASES = 6
    # ret.extend(random.randrange(lo, hi) for _ in range(NUM_RANDOM_CASES))

    return ret


def _cases_for_decimal(typ):
    lo, hi = typ.ast_bounds

    ret = [Decimal(i) for i in [-1, 0, 1]]
    ret.extend([lo - 1, lo, lo + 1, hi - 1, hi, hi + 1])

    ret.extend(
        [lo - DECIMAL_EPSILON, lo + DECIMAL_EPSILON, hi - DECIMAL_EPSILON, hi + DECIMAL_EPSILON]
    )

    # random cases cause reproducibility issues. TODO fixme
    # (use int values because randrange can't generate fractional decimals)
    # int_lo, int_hi = info.bounds  # e.g. -(2**167)
    # NUM_RANDOM_CASES = 10  # more than int, just for paranoia's sake
    # DIVISOR = info.divisor
    # ret.extend(random.randrange(int_lo, int_hi) / DIVISOR for _ in range(NUM_RANDOM_CASES))

    return ret


def _cases_for_address(_typ):
    cases = _filter_cases(_cases_for_int(UINT160_T), UINT160_T)
    return [_py_convert(c, UINT160_T, AddressT()) for c in cases]


def _cases_for_bool(_typ):
    return [True, False]


def _cases_for_bytes(typ):
    # reuse the cases for the equivalent int type
    equiv_int_type = UINT(typ.m_bits)
    cases = _filter_cases(_cases_for_int(equiv_int_type), equiv_int_type)
    return [_py_convert(c, equiv_int_type, typ) for c in cases]


def _cases_for_Bytes(typ):
    ret = []
    # would not need this if we tested all Bytes[1]...Bytes[32] types.
    for i in range(32):
        ret.extend(_cases_for_bytes(BytesM_T(i + 1)))

    ret.append(b"")
    return uniq(ret)


def _cases_for_String(typ):
    ret = []
    # would not need this if we tested all Bytes[1]...Bytes[32] types.
    for i in range(32):
        ret.extend([str(c, "utf-8") for c in _cases_for_bytes(BytesM_T(i + 1))])
    ret.append("")
    return uniq(ret)


# generate all cases of interest for a type, potentially including invalid cases
def interesting_cases_for_type(typ):
    if isinstance(typ, IntegerT):
        return _cases_for_int(typ)
    if isinstance(typ, DecimalT):
        return _cases_for_decimal(typ)
    if isinstance(typ, BytesM_T):
        return _cases_for_bytes(typ)
    if isinstance(typ, BytesT):
        return _cases_for_Bytes(typ)
    if isinstance(typ, StringT):
        return _cases_for_String(typ)
    if isinstance(typ, BoolT):
        return _cases_for_bool(typ)
    if isinstance(typ, AddressT):
        return _cases_for_address(typ)


def _filter_cases(cases, i_typ):
    cases = uniq(cases)

    def _in_bounds(c):
        try:
            return _py_convert(c, i_typ, i_typ) is not None
        except eth.codecs.abi.exceptions.EncodeError:
            return False

    return [c for c in cases if _in_bounds(c)]


class _PadDirection(enum.Enum):
    Left = enum.auto()
    Right = enum.auto()


def _padding_direction(typ):
    if isinstance(typ, (BytesM_T, StringT, BytesT)):
        return _PadDirection.Right
    return _PadDirection.Left


# TODO this could be a function in vyper.builtins._convert
# which implements literal folding and also serves as a reference/spec
def _padconvert(val_bits, direction, n, padding_byte=None):
    """
    Takes the ABI representation of a value, and convert the padding if needed.
    If fill_zeroes is false, the two halves of the bytestring are just swapped
    and the dirty bytes remain dirty. If fill_zeroes is true, the padding
    bytes get set to 0
    """
    assert len(val_bits) == 32

    # convert left-padded to right-padded
    if direction == _PadDirection.Right:
        tail = val_bits[:-n]
        if padding_byte is not None:
            tail = padding_byte * len(tail)
        return val_bits[-n:] + tail

    # right- to left- padded
    if direction == _PadDirection.Left:
        head = val_bits[n:]
        if padding_byte is not None:
            head = padding_byte * len(head)
        return head + val_bits[:n]


def _from_bits(val_bits, o_typ):
    # o_typ: the type to convert to
    try:
        ret = abi.decode(o_typ.abi_type.selector_name(), val_bits)
        if isinstance(o_typ, DecimalT):
            return Decimal(ret) / o_typ.divisor
        return ret
    except eth.codecs.abi.exceptions.DecodeError:
        raise _OutOfBounds() from None


def _to_bits(val, i_typ):
    # i_typ: the type to convert from
    if isinstance(i_typ, DecimalT):
        val = val * i_typ.divisor
        assert math.ceil(val) == math.floor(val)
        val = int(val)
    return abi.encode(i_typ.abi_type.selector_name(), val)


def _signextend(val_bytes, bits):
    as_uint = int.from_bytes(val_bytes, byteorder="big")

    as_sint = unsigned_to_signed(as_uint, bits)

    return (as_sint % 2**256).to_bytes(32, byteorder="big")


def _convert_int_to_int(val, o_typ):
    lo, hi = o_typ.int_bounds
    if not lo <= val <= hi:
        return None
    return val


def _convert_decimal_to_int(val, o_typ):
    # note special behavior for decimal: catch OOB before truncation.
    lo, hi = o_typ.int_bounds
    if not lo <= val <= hi:
        return None

    return round_towards_zero(val)


def _convert_int_to_decimal(val, o_typ):
    ret = Decimal(val)
    lo, hi = o_typ.ast_bounds

    if not lo <= ret <= hi:
        return None

    return ret


def _py_convert(val, i_typ, o_typ):
    """
    Perform conversion on the Python representation of a Vyper value.
    Returns None if the conversion is invalid (i.e., would revert in Vyper)
    """

    if isinstance(i_typ, IntegerT) and isinstance(o_typ, IntegerT):
        return _convert_int_to_int(val, o_typ)

    if isinstance(i_typ, DecimalT) and isinstance(o_typ, IntegerT):
        return _convert_decimal_to_int(val, o_typ)

    if isinstance(i_typ, (BoolT, IntegerT)) and isinstance(o_typ, DecimalT):
        # Note: Decimal(True) == Decimal("1")
        return _convert_int_to_decimal(val, o_typ)

    val_bits = _to_bits(val, i_typ)

    if isinstance(i_typ, (BytesT, StringT)):
        val_bits = val_bits[32:]

    if _padding_direction(i_typ) != _padding_direction(o_typ):
        # subtle! the padding conversion follows the bytes argument
        if isinstance(i_typ, (BytesM_T, BytesT)):
            n = bytes_of_type(i_typ)
            padding_byte = None
        else:
            # output type is bytes
            n = bytes_of_type(o_typ)
            padding_byte = b"\x00"

        val_bits = _padconvert(val_bits, _padding_direction(o_typ), n, padding_byte)

    if getattr(o_typ, "is_signed", False) and isinstance(i_typ, BytesM_T):
        n_bits = _bits_of_type(i_typ)
        val_bits = _signextend(val_bits, n_bits)

    try:
        if isinstance(o_typ, BoolT):
            return _from_bits(val_bits, UINT256_T) != 0

        ret = _from_bits(val_bits, o_typ)

        if isinstance(o_typ, AddressT):
            return checksum_encode(ret)
        return ret

    except _OutOfBounds:
        return None


# the matrix of all type pairs
def all_pairs():
    return sorted(itertools.product(BASE_TYPES, BASE_TYPES))


# pairs which can compile
def convertible_pairs():
    return [(i, o) for (i, o) in all_pairs() if can_convert(i, o)]


# pairs which shouldn't even compile
def non_convertible_pairs():
    return [(i, o) for (i, o) in all_pairs() if not can_convert(i, o) and i != o]


# _CASES_CACHE = {}


def cases_for_pair(i_typ, o_typ):
    """
    Helper function to generate all cases for pair
    """
    # if (i_typ, o_typ) in _CASES_CACHE:
    #    # cache the cases for reproducibility, to ensure test_passing_cases and test_failing_cases
    #    # test exactly the two halves of the produced cases.
    #    return _CASES_CACHE[(i_typ, o_typ)]

    cases = interesting_cases_for_type(i_typ)
    # only return cases which are valid for the input type
    cases = _filter_cases(cases, i_typ)

    for c in interesting_cases_for_type(o_typ):
        # convert back into i_typ
        try:
            c = _py_convert(c, o_typ, i_typ)
            if c is not None:
                cases.append(c)
        except eth.codecs.abi.exceptions.EncodeError:
            pass

    # _CASES_CACHE[(i_typ, o_typ)] = cases

    return cases


def generate_passing_cases():
    ret = []
    for i_typ, o_typ in convertible_pairs():
        cases = cases_for_pair(i_typ, o_typ)
        for c in cases:
            # only add convertible cases
            if _py_convert(c, i_typ, o_typ) is not None:
                ret.append((i_typ, o_typ, c))
    return sorted(ret)


def generate_reverting_cases():
    ret = []
    for i_typ, o_typ in convertible_pairs():
        cases = cases_for_pair(i_typ, o_typ)
        for c in cases:
            if _py_convert(c, i_typ, o_typ) is None:
                ret.append((i_typ, o_typ, c))
    return sorted(ret)


def _vyper_literal(val, typ):
    if isinstance(typ, BytesM_T):
        return "0x" + val.hex()
    if isinstance(typ, DecimalT):
        tmp = val
        val = quantize(val)
        assert tmp == val
    return str(val)


@pytest.mark.parametrize("i_typ,o_typ,val", generate_passing_cases())
@pytest.mark.fuzzing
def test_convert_passing(get_contract, assert_compile_failed, i_typ, o_typ, val):
    expected_val = _py_convert(val, i_typ, o_typ)

    if isinstance(o_typ, DecimalT):
        expected_val = decimal_to_int(expected_val)

    input_val = val
    if isinstance(i_typ, DecimalT):
        input_val = decimal_to_int(val)

    contract_1 = f"""
@external
def test_convert() -> {o_typ}:
    return convert({_vyper_literal(val, i_typ)}, {o_typ})
    """

    skip_c1 = False

    # Skip bytes20 literals when there is ambiguity with `address` since address takes precedence.
    # generally happens when there are only digits in the literal.
    if i_typ == BYTES20_T and is_checksum_encoded(_vyper_literal(val, BYTES20_T)):
        skip_c1 = True

    # typechecker inference borked, ambiguity with bytes20
    if isinstance(i_typ, AddressT) and o_typ == BYTES20_T and val == val.lower():
        skip_c1 = True

    if not skip_c1:
        c1 = get_contract(contract_1)
        assert c1.test_convert() == expected_val

    contract_2 = f"""
@external
def test_input_convert(x: {i_typ}) -> {o_typ}:
    return convert(x, {o_typ})
    """

    c2 = get_contract(contract_2)
    assert c2.test_input_convert(input_val) == expected_val

    contract_3 = f"""
bar: {i_typ}

@external
def test_state_variable_convert() -> {o_typ}:
    self.bar = {_vyper_literal(val, i_typ)}
    return convert(self.bar, {o_typ})
    """

    c3 = get_contract(contract_3)
    assert c3.test_state_variable_convert() == expected_val

    contract_4 = f"""
@external
def test_memory_variable_convert(x: {i_typ}) -> {o_typ}:
    y: {i_typ} = x
    return convert(y, {o_typ})
    """

    c4 = get_contract(contract_4)
    assert c4.test_memory_variable_convert(input_val) == expected_val


@pytest.mark.parametrize("typ", ["uint8", "int128", "int256", "uint256"])
@pytest.mark.parametrize("val", [1, 2, 2**128, 2**256 - 1, 2**256 - 2])
def test_flag_conversion(get_contract, assert_compile_failed, val, typ):
    roles = "\n    ".join([f"ROLE_{i}" for i in range(256)])
    contract = f"""
flag Roles:
    {roles}

@external
def foo(a: Roles) -> {typ}:
    return convert(a, {typ})

@external
def bar(a: uint256) -> Roles:
    return convert(a, Roles)
    """
    if typ == "uint256":
        c = get_contract(contract)
        assert c.foo(val) == val
        assert c.bar(val) == val
    else:
        assert_compile_failed(lambda: get_contract(contract), TypeMismatch)


@pytest.mark.parametrize("typ", ["uint8", "int128", "int256", "uint256"])
@pytest.mark.parametrize("val", [1, 2, 3, 4, 2**128, 2**256 - 1, 2**256 - 2])
def test_flag_conversion_2(get_contract, assert_compile_failed, tx_failed, val, typ):
    contract = f"""
flag Status:
    STARTED
    PAUSED
    STOPPED

@external
def foo(a: {typ}) -> Status:
    return convert(a, Status)
    """
    if typ == "uint256":
        c = get_contract(contract)
        lo, hi = int_bounds(signed=False, bits=3)
        if lo <= val <= hi:
            assert c.foo(val) == val
        else:
            with tx_failed():
                c.foo(val)
    else:
        assert_compile_failed(lambda: get_contract(contract), TypeMismatch)


# uint256 conversion is currently valid due to type inference on literals
# not quite working yet
same_type_conversion_blocked = sorted(TEST_TYPES - {UINT256_T})


@pytest.mark.parametrize("typ", same_type_conversion_blocked)
def test_same_type_conversion_blocked(get_contract, assert_compile_failed, typ):
    code = f"""
@external
def foo(x: {typ}) -> {typ}:
    return convert(x, {typ})
    """
    assert_compile_failed(lambda: get_contract(code), InvalidType)


@pytest.mark.parametrize("i_typ,o_typ", non_convertible_pairs())
def test_type_conversion_blocked(get_contract, assert_compile_failed, i_typ, o_typ):
    code = f"""
@external
def foo(x: {i_typ}) -> {o_typ}:
    return convert(x, {o_typ})
    """
    assert_compile_failed(lambda: get_contract(code), TypeMismatch)


@pytest.mark.parametrize("typ", sorted(BASE_TYPES))
def test_bytes_too_large_cases(typ):
    code_1 = f"""
@external
def foo(x: Bytes[33]) -> {typ}:
    return convert(x, {typ})
    """
    with pytest.raises(TypeMismatch):
        compile_code(code_1)

    bytes_33 = b"1" * 33
    code_2 = f"""
@external
def foo() -> {typ}:
    return convert({bytes_33}, {typ})
    """
    with pytest.raises(TypeMismatch):
        compile_code(code_2)


@pytest.mark.parametrize("cls1,cls2", itertools.product((StringT, BytesT), (StringT, BytesT)))
def test_bytestring_conversions(cls1, cls2, get_contract, tx_failed):
    typ1 = cls1(33)
    typ2 = cls2(32)

    def bytestring(cls, string):
        if cls == BytesT:
            return string.encode("utf-8")
        return string

    code_1 = f"""
@external
def foo(x: {typ1}) -> {typ2}:
    return convert(x, {typ2})
    """
    c = get_contract(code_1)

    for i in range(33):  # inclusive 32
        s = "1" * i
        arg = bytestring(cls1, s)
        out = bytestring(cls2, s)
        assert c.foo(arg) == out

    with tx_failed():
        # TODO: sanity check it is convert which is reverting, not arg clamping
        c.foo(bytestring(cls1, "1" * 33))

    code_2_template = """
@external
def foo() -> {typ}:
    return convert({arg}, {typ})
    """

    # test literals
    for i in range(33):  # inclusive 32
        s = "1" * i
        arg = bytestring(cls1, s)
        out = bytestring(cls2, s)
        code = code_2_template.format(typ=typ2, arg=repr(arg))
        if cls1 == cls2:  # ex.: can't convert "" to String[32]
            with pytest.raises(InvalidType):
                compile_code(code)
        else:
            c = get_contract(code)
            assert c.foo() == out

    failing_code = code_2_template.format(typ=typ2, arg=bytestring(cls1, "1" * 33))
    with pytest.raises(TypeMismatch):
        compile_code(failing_code)


@pytest.mark.parametrize("n", range(1, 33))
def test_Bytes_to_bytes(get_contract, n: int):
    t_bytes = f"bytes{n}"
    t_Bytes = f"Bytes[{n}]"

    test_data = b"\xff" * n

    code1 = f"""
@external
def foo() -> {t_bytes}:
    x: {t_Bytes} = {test_data}
    return convert(x, {t_bytes})
    """
    c1 = get_contract(code1)
    assert c1.foo() == test_data

    code2 = f"""
bar: {t_Bytes}
@external
def foo() -> {t_bytes}:
    self.bar = {test_data}
    return convert(self.bar, {t_bytes})
    """
    c2 = get_contract(code2)
    assert c2.foo() == test_data


@pytest.mark.parametrize("i_typ,o_typ,val", generate_reverting_cases())
@pytest.mark.fuzzing
def test_conversion_failures(get_contract, assert_compile_failed, tx_failed, i_typ, o_typ, val):
    """
    Test multiple contracts and check for a specific exception.
    If no exception is provided, a runtime revert is expected (e.g. clamping).
    """
    contract_1 = f"""
@external
def foo() -> {o_typ}:
    return convert({_vyper_literal(val, i_typ)}, {o_typ})
    """

    c1_exception = InvalidLiteral

    if isinstance(i_typ, IntegerT) and isinstance(o_typ, BytesM_T):
        # integer literals get upcasted to uint256 / int256 types, so the convert
        # will not compile unless it is bytes32
        if o_typ != BYTES32_T:
            c1_exception = TypeMismatch

    # compile-time folding not implemented for these:
    skip_c1 = False
    # if isinstance(o_typ, IntegerT.signeds()) and isinstance(i_typ, Address()):
    #    skip_c1 = True

    if isinstance(o_typ, BytesM_T):
        skip_c1 = True

    # if o_typ in (AddressT(), BYTES20_T):
    #    skip_c1 = True

    if not skip_c1:
        assert_compile_failed(lambda: get_contract(contract_1), c1_exception)

    contract_2 = f"""
@external
def foo():
    bar: {i_typ} = {_vyper_literal(val, i_typ)}
    foobar: {o_typ} = convert(bar, {o_typ})
    """

    c2 = get_contract(contract_2)
    with tx_failed():
        c2.foo()

    contract_3 = f"""
@external
def foo(bar: {i_typ}) -> {o_typ}:
    return convert(bar, {o_typ})
    """

    c3 = get_contract(contract_3)
    input_val = val
    if isinstance(i_typ, DecimalT):
        input_val = decimal_to_int(input_val)
    with tx_failed():
        c3.foo(input_val)



import binascii
import contextlib
import decimal
import enum
import functools
import hashlib
import sys
import time
import traceback
import warnings
from typing import Generic, List, TypeVar, Union

from vyper.exceptions import CompilerPanic, DecimalOverrideException, InvalidLiteral, VyperException

_T = TypeVar("_T")


class OrderedSet(Generic[_T]):
    """
    a minimal "ordered set" class. this is needed in some places
    because, while dict guarantees you can recover insertion order
    vanilla sets do not.
    no attempt is made to fully implement the set API, will add
    functionality as needed.
    """

    def __init__(self, iterable=None):
        self._data = dict()
        if iterable is not None:
            self.update(iterable)

    def __repr__(self):
        keys = ", ".join(repr(k) for k in self)
        return f"{{{keys}}}"

    def __iter__(self):
        return iter(self._data)

    def __contains__(self, item):
        return self._data.__contains__(item)

    def __len__(self):
        return len(self._data)

    def first(self):
        return next(iter(self))

    def add(self, item: _T) -> None:
        self._data[item] = None

    def remove(self, item: _T) -> None:
        del self._data[item]

    def drop(self, item: _T):
        # friendly version of remove
        self._data.pop(item, None)

    def dropmany(self, iterable):
        for item in iterable:
            self._data.pop(item, None)

    def difference(self, other):
        ret = self.copy()
        ret.dropmany(other)
        return ret

    def update(self, other):
        # CMC 2024-03-22 for some reason, this is faster than dict.update?
        # (maybe size dependent)
        for item in other:
            self._data[item] = None

    def union(self, other):
        return self | other

    def __ior__(self, other):
        self.update(other)
        return self

    def __or__(self, other):
        ret = self.copy()
        ret.update(other)
        return ret

    def __eq__(self, other):
        return self._data == other._data

    def copy(self):
        cls = self.__class__
        ret = cls.__new__(cls)
        ret._data = self._data.copy()
        return ret

    @classmethod
    def intersection(cls, *sets):
        if len(sets) == 0:
            raise ValueError("undefined: intersection of no sets")

        ret = sets[0].copy()
        for e in sets[0]:
            if any(e not in s for s in sets[1:]):
                ret.remove(e)
        return ret


class StringEnum(enum.Enum):
    # Must be first, or else won't work, specifies what .value is
    @staticmethod
    def _generate_next_value_(name, start, count, last_values):
        return name.lower()

    # Override ValueError with our own internal exception
    @classmethod
    def _missing_(cls, value):
        raise CompilerPanic(f"{value} is not a valid {cls.__name__}")

    @classmethod
    def is_valid_value(cls, value: str) -> bool:
        return value in set(o.value for o in cls)

    @classmethod
    def options(cls) -> List["StringEnum"]:
        return list(cls)

    @classmethod
    def values(cls) -> List[str]:
        return [v.value for v in cls.options()]

    # Comparison operations
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            raise CompilerPanic(f"bad comparison: ({type(other)}, {type(self)})")
        return self is other

    # Python normally does __ne__(other) ==> not self.__eq__(other)

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            raise CompilerPanic(f"bad comparison: ({type(other)}, {type(self)})")
        options = self.__class__.options()
        return options.index(self) < options.index(other)  # type: ignore

    def __le__(self, other: object) -> bool:
        return self.__eq__(other) or self.__lt__(other)

    def __gt__(self, other: object) -> bool:
        return not self.__le__(other)

    def __ge__(self, other: object) -> bool:
        return not self.__lt__(other)

    def __str__(self) -> str:
        return self.value

    def __hash__(self) -> int:
        # let `dataclass` know that this class is not mutable
        return super().__hash__()


class DecimalContextOverride(decimal.Context):
    def __setattr__(self, name, value):
        if name == "prec":
            if value < 78:
                # definitely don't want this to happen
                raise DecimalOverrideException("Overriding decimal precision disabled")
            elif value > 78:
                # not sure it's incorrect, might not be end of the world
                warnings.warn(
                    "Changing decimals precision could have unintended side effects!", stacklevel=2
                )
            # else: no-op, is ok

        super().__setattr__(name, value)


decimal.setcontext(DecimalContextOverride(prec=78))


try:
    from Crypto.Hash import keccak  # type: ignore

    keccak256 = lambda x: keccak.new(digest_bits=256, data=x).digest()  # noqa: E731
except ImportError:
    import sha3 as _sha3

    keccak256 = lambda x: _sha3.sha3_256(x).digest()  # noqa: E731


@functools.lru_cache(maxsize=512)
def sha256sum(s: str) -> str:
    return hashlib.sha256(s.encode("utf-8")).digest().hex()


def get_long_version():
    from vyper import __long_version__

    return __long_version__


# Converts four bytes to an integer
def fourbytes_to_int(inp):
    return (inp[0] << 24) + (inp[1] << 16) + (inp[2] << 8) + inp[3]


# Converts an integer to four bytes
def int_to_fourbytes(n: int) -> bytes:
    assert n < 2**32
    return n.to_bytes(4, byteorder="big")


def signed_to_unsigned(int_, bits, strict=False):
    """
    Reinterpret a signed integer with n bits as an unsigned integer.
    The implementation is unforgiving in that it assumes the input is in
    bounds for int<bits>, in order to fail more loudly (and not hide
    errors in modular reasoning in consumers of this function).
    """
    if strict:
        lo, hi = int_bounds(signed=True, bits=bits)
        assert lo <= int_ <= hi
    if int_ < 0:
        return int_ + 2**bits
    return int_


def unsigned_to_signed(int_, bits, strict=False):
    """
    Reinterpret an unsigned integer with n bits as a signed integer.
    The implementation is unforgiving in that it assumes the input is in
    bounds for uint<bits>, in order to fail more loudly (and not hide
    errors in modular reasoning in consumers of this function).
    """
    if strict:
        lo, hi = int_bounds(signed=False, bits=bits)
        assert lo <= int_ <= hi
    if int_ > (2 ** (bits - 1)) - 1:
        return int_ - (2**bits)
    return int_


def is_power_of_two(n: int) -> bool:
    # busted for ints wider than 53 bits:
    # t = math.log(n, 2)
    # return math.ceil(t) == math.floor(t)
    return n != 0 and ((n & (n - 1)) == 0)


# https://stackoverflow.com/a/71122440/
def int_log2(n: int) -> int:
    return n.bit_length() - 1


# utility function for debugging purposes
def trace(n=5, out=sys.stderr):
    print("BEGIN TRACE", file=out)
    for x in list(traceback.format_stack())[-n:]:
        print(x.strip(), file=out)
    print("END TRACE", file=out)


# print a warning
def vyper_warn(msg, node=None):
    if node is not None:
        # use VyperException for its formatting abilities
        msg = str(VyperException(msg, node))
    warnings.warn(msg, stacklevel=2)


# converts a signature like Func(bool,uint256,address) to its 4 byte method ID
# TODO replace manual calculations in codebase with this
def method_id_int(method_sig: str) -> int:
    method_id_bytes = method_id(method_sig)
    return fourbytes_to_int(method_id_bytes)


def method_id(method_str: str) -> bytes:
    return keccak256(bytes(method_str, "utf-8"))[:4]


def round_towards_zero(d: decimal.Decimal) -> int:
    # TODO double check if this can just be int(d)
    # (but either way keep this util function bc it's easier at a glance
    # to understand what round_towards_zero() does instead of int())
    return int(d.to_integral_exact(decimal.ROUND_DOWN))


# Converts string to bytes
def string_to_bytes(str):
    bytez = b""
    for c in str:
        if ord(c) >= 256:
            raise InvalidLiteral(f"Cannot insert special character {c} into byte array")
        bytez += bytes([ord(c)])
    bytez_length = len(bytez)
    return bytez, bytez_length


# Converts a provided hex string to an integer
def hex_to_int(inp):
    if inp[:2] == "0x":
        inp = inp[2:]
    return bytes_to_int(binascii.unhexlify(inp))


# Converts bytes to an integer
def bytes_to_int(bytez):
    o = 0
    for b in bytez:
        o = o * 256 + b
    return o


def is_checksum_encoded(addr):
    return addr == checksum_encode(addr)


# Encodes an address using ethereum's checksum scheme
def checksum_encode(addr):  # Expects an input of the form 0x<40 hex chars>
    assert addr[:2] == "0x" and len(addr) == 42, addr
    o = ""
    v = bytes_to_int(keccak256(addr[2:].lower().encode("utf-8")))
    for i, c in enumerate(addr[2:]):
        if c in "0123456789":
            o += c
        else:
            o += c.upper() if (v & (2 ** (255 - 4 * i))) else c.lower()
    return "0x" + o


# Returns lowest multiple of 32 >= the input
def ceil32(x):
    return x if x % 32 == 0 else x + 32 - (x % 32)


# Calculates amount of gas needed for memory expansion
def calc_mem_gas(memsize):
    return (memsize // 32) * 3 + (memsize // 32) ** 2 // 512


# Specific gas usage
GAS_IDENTITY = 15
GAS_IDENTITYWORD = 3
GAS_COPY_WORD = 3  # i.e., W_copy from YP

# A decimal value can store multiples of 1/DECIMAL_DIVISOR
MAX_DECIMAL_PLACES = 10
DECIMAL_DIVISOR = 10**MAX_DECIMAL_PLACES
DECIMAL_EPSILON = decimal.Decimal(1) / DECIMAL_DIVISOR


def int_bounds(signed, bits):
    """
    calculate the bounds on an integer type
    ex. int_bounds(True, 8) -> (-128, 127)
        int_bounds(False, 8) -> (0, 255)
    """
    if signed:
        return -(2 ** (bits - 1)), (2 ** (bits - 1)) - 1
    return 0, (2**bits) - 1


# e.g. -1 -> -(2**256 - 1)
def evm_twos_complement(x: int) -> int:
    # return ((o + 2 ** 255) % 2 ** 256) - 2 ** 255
    return ((2**256 - 1) ^ x) + 1


# EVM div semantics as a python function
def evm_div(x, y):
    if y == 0:
        return 0
    # NOTE: should be same as: round_towards_zero(Decimal(x)/Decimal(y))
    sign = -1 if (x * y) < 0 else 1
    return sign * (abs(x) // abs(y))  # adapted from py-evm


# EVM mod semantics as a python function
def evm_mod(x, y):
    if y == 0:
        return 0

    sign = -1 if x < 0 else 1
    return sign * (abs(x) % abs(y))  # adapted from py-evm


# EVM pow which wraps instead of hanging on "large" numbers
# (which can generated, for ex. in the unevaluated branch of the Shift builtin)
def evm_pow(x, y):
    assert x >= 0 and y >= 0
    return pow(x, y, 2**256)


# memory used for system purposes, not for variables
class MemoryPositions:
    FREE_VAR_SPACE = 0
    FREE_VAR_SPACE2 = 32
    RESERVED_MEMORY = 64


# Sizes of different data types. Used to clamp types.
class SizeLimits:
    MAX_INT128 = 2**127 - 1
    MIN_INT128 = -(2**127)
    MAX_INT256 = 2**255 - 1
    MIN_INT256 = -(2**255)
    MAXDECIMAL = 2**167 - 1  # maxdecimal as EVM value
    MINDECIMAL = -(2**167)  # mindecimal as EVM value
    # min decimal allowed as Python value
    MIN_AST_DECIMAL = -decimal.Decimal(2**167) / DECIMAL_DIVISOR
    # max decimal allowed as Python value
    MAX_AST_DECIMAL = decimal.Decimal(2**167 - 1) / DECIMAL_DIVISOR
    MAX_UINT8 = 2**8 - 1
    MAX_UINT256 = 2**256 - 1
    CEILING_UINT256 = 2**256


def quantize(d: decimal.Decimal, places=MAX_DECIMAL_PLACES, rounding_mode=decimal.ROUND_DOWN):
    quantizer = decimal.Decimal(f"{1:0.{places}f}")
    return d.quantize(quantizer, rounding_mode)


# List of valid IR macros.
# TODO move this somewhere else, like ir_node.py
VALID_IR_MACROS = {
    "assert",
    "break",
    "iload",
    "istore",
    "dload",
    "dloadbytes",
    "ceil32",
    "continue",
    "debugger",
    "ge",
    "if",
    "select",
    "le",
    "deploy",
    "ne",
    "pass",
    "repeat",
    "seq",
    "set",
    "sge",
    "sha3_32",
    "sha3_64",
    "sle",
    "with",
    "label",
    "goto",
    "djump",  # "dynamic jump", i.e. constrained, multi-destination jump
    "~extcode",
    "~selfcode",
    "~calldata",
    "~empty",
    "var_list",
}


EIP_170_LIMIT = 0x6000  # 24kb
ERC5202_PREFIX = b"\xFE\x71\x00"  # default prefix from ERC-5202

SHA3_BASE = 30
SHA3_PER_WORD = 6


def indent(text: str, indent_chars: Union[str, List[str]] = " ", level: int = 1) -> str:
    """
    Indent lines of text in the string ``text`` using the indentation
    character(s) given in ``indent_chars`` ``level`` times.

    :param text: A string containing the lines of text to be indented.
    :param level: The number of times to indent lines in ``text``.
    :param indent_chars: The characters to use for indentation.  If a string,
        uses repetitions of that string for indentation.  If a list of strings,
        uses repetitions of each string to indent each line.

    :return: The indented text.
    """
    text_lines = text.splitlines(keepends=True)

    if isinstance(indent_chars, str):
        indented_lines = [indent_chars * level + line for line in text_lines]
    elif isinstance(indent_chars, list):
        if len(indent_chars) != len(text_lines):
            raise ValueError("Must provide indentation chars for each line")

        indented_lines = [ind * level + line for ind, line in zip(indent_chars, text_lines)]
    else:
        raise ValueError("Unrecognized indentation characters value")

    return "".join(indented_lines)


@contextlib.contextmanager
def timeit(msg):
    start_time = time.perf_counter()
    yield
    end_time = time.perf_counter()
    total_time = end_time - start_time
    print(f"{msg}: Took {total_time:.4f} seconds")


@contextlib.contextmanager
def timer(msg):
    t0 = time.time()
    yield
    t1 = time.time()
    print(f"{msg} took {t1 - t0}s")


def annotate_source_code(
    source_code: str,
    lineno: int,
    col_offset: int = None,
    context_lines: int = 0,
    line_numbers: bool = False,
) -> str:
    """
    Annotate the location specified by ``lineno`` and ``col_offset`` in the
    source code given by ``source_code`` with a location marker and optional
    line numbers and context lines.

    :param source_code: The source code containing the source location.
    :param lineno: The 1-indexed line number of the source location.
    :param col_offset: The 0-indexed column offset of the source location.
    :param context_lines: The number of contextual lines to include above and
        below the source location.
    :param line_numbers: If true, line numbers are included in the location
        representation.

    :return: A string containing the annotated source code location.
    """
    if lineno is None:
        return ""

    source_lines = source_code.splitlines(keepends=True)
    if lineno < 1 or lineno > len(source_lines):
        raise ValueError("Line number is out of range")

    line_offset = lineno - 1
    start_offset = max(0, line_offset - context_lines)
    end_offset = min(len(source_lines), line_offset + context_lines + 1)

    line_repr = source_lines[line_offset]
    if "\n" not in line_repr[-2:]:  # Handle certain edge cases
        line_repr += "\n"
    if col_offset is None:
        mark_repr = ""
    else:
        mark_repr = "-" * col_offset + "^" + "\n"

    before_lines = "".join(source_lines[start_offset:line_offset])
    after_lines = "".join(source_lines[line_offset + 1 : end_offset])
    location_repr = "".join((before_lines, line_repr, mark_repr, after_lines))

    if line_numbers:
        # Create line numbers
        lineno_reprs = [f"{i} " for i in range(start_offset + 1, end_offset + 1)]

        # Highlight line identified by `lineno`
        local_line_off = line_offset - start_offset
        lineno_reprs[local_line_off] = "---> " + lineno_reprs[local_line_off]

        # Calculate width of widest line no
        max_len = max(len(i) for i in lineno_reprs)

        # Justify all line nos according to this width
        justified_reprs = [i.rjust(max_len) for i in lineno_reprs]
        if col_offset is not None:
            justified_reprs.insert(local_line_off + 1, "-" * max_len)

        location_repr = indent(location_repr, indent_chars=justified_reprs)

    # Ensure no trailing whitespace and trailing blank lines are only included
    # if they are part of the source code
    if col_offset is None:
        # Number of lines doesn't include column marker line
        num_lines = end_offset - start_offset
    else:
        num_lines = end_offset - start_offset + 1

    cleanup_lines = [line.rstrip() for line in location_repr.splitlines()]
    cleanup_lines += [""] * (num_lines - len(cleanup_lines))

    return "\n".join(cleanup_lines)



from vyper.utils import OrderedSet
from vyper.venom.analysis.dfg import DFGAnalysis
from vyper.venom.basicblock import BB_TERMINATORS, IRBasicBlock, IRInstruction, IRVariable
from vyper.venom.function import IRFunction
from vyper.venom.passes.base_pass import IRPass


class DFTPass(IRPass):
    function: IRFunction
    inst_order: dict[IRInstruction, int]
    inst_order_num: int

    def _process_instruction_r(self, bb: IRBasicBlock, inst: IRInstruction, offset: int = 0):
        for op in inst.get_outputs():
            assert isinstance(op, IRVariable), f"expected variable, got {op}"
            uses = self.dfg.get_uses(op)

            for uses_this in uses:
                if uses_this.parent != inst.parent or uses_this.fence_id != inst.fence_id:
                    # don't reorder across basic block or fence boundaries
                    continue

                # if the instruction is a terminator, we need to place
                # it at the end of the basic block
                # along with all the instructions that "lead" to it
                self._process_instruction_r(bb, uses_this, offset)

        if inst in self.visited_instructions:
            return
        self.visited_instructions.add(inst)
        self.inst_order_num += 1

        if inst.opcode in BB_TERMINATORS:
            offset = len(bb.instructions)

        if inst.opcode == "phi":
            # phi instructions stay at the beginning of the basic block
            # and no input processing is needed
            # bb.instructions.append(inst)
            self.inst_order[inst] = 0
            return

        for op in inst.get_inputs():
            target = self.dfg.get_producing_instruction(op)
            assert target is not None, f"no producing instruction for {op}"
            if target.parent != inst.parent or target.fence_id != inst.fence_id:
                # don't reorder across basic block or fence boundaries
                continue
            self._process_instruction_r(bb, target, offset)

        self.inst_order[inst] = self.inst_order_num + offset

    def _process_basic_block(self, bb: IRBasicBlock) -> None:
        self.function.append_basic_block(bb)

        for inst in bb.instructions:
            inst.fence_id = self.fence_id
            if inst.volatile:
                self.fence_id += 1

        # We go throught the instructions and calculate the order in which they should be executed
        # based on the data flow graph. This order is stored in the inst_order dictionary.
        # We then sort the instructions based on this order.
        self.inst_order = {}
        self.inst_order_num = 0
        for inst in bb.instructions:
            self._process_instruction_r(bb, inst)

        bb.instructions.sort(key=lambda x: self.inst_order[x])

    def run_pass(self) -> None:
        self.dfg = self.analyses_cache.request_analysis(DFGAnalysis)

        self.fence_id = 0
        self.visited_instructions: OrderedSet[IRInstruction] = OrderedSet()

        basic_blocks = list(self.function.get_basic_blocks())

        self.function.clear_basic_blocks()
        for bb in basic_blocks:
            self._process_basic_block(bb)



import ast as python_ast
from typing import Any, Optional, Union

from . import nodes, validation
from .natspec import parse_natspec as parse_natspec
from .nodes import *
from .parse import parse_to_ast as parse_to_ast
from .utils import ast_to_dict as ast_to_dict
